#!python3.9

from . import EventMessage
from xml.etree.ElementTree import Element

class LocationSelectEventMessage(EventMessage):

    # handler key 
    key:str = 'event.location_select'

    def __init__(self, to_username: str, from_username: str, agent_id: str, event: str, loc_x:str, loc_y:str, scale:str, label:str, poi_name:str, event_key: str = '', **kwargs) -> None:
        super().__init__(to_username, from_username, agent_id, event, event_key=event_key, **kwargs)
        # location x
        self.location_x:str = loc_x
        # location y 
        self.location_y:str = loc_y
        # scale 
        self.scale:str = scale
        # label for the position 
        self.label:str = label
        self.poiname:str = poi_name

    @classmethod
    def from_xml(cls, xml_tree: Element):
        '''Parse location_select event message from XML object'''
        to_username:str = xml_tree.find('ToUserName').text
        from_username:str = xml_tree.find('FromUserName').text
        create_time:str = xml_tree.find('CreateTime').text
        event:str = xml_tree.find('Event').text
        event_key:str = xml_tree.find('EventKey').text
        loc_x:str = xml_tree.find('SendLocationInfo/Location_X').text
        loc_y:str = xml_tree.find('SendLocationInfo/Location_Y').text
        scale:str = xml_tree.find('SendLocationInfo/Scale').text
        label:str = xml_tree.find('SendLocationInfo/Label').text
        poi_name:str = xml_tree.find('SendLocationInfo/Poiname').text
        # msg_id:str = xml_tree.find('MsgId').text
        agent_id:str = xml_tree.find('AgentID').text

        return cls(to_username, from_username, agent_id, event, loc_x, loc_y, scale, label, poi_name, event_key=event_key, create_time=create_time)


#!python3.9

from xml.etree.ElementTree import Element
from . import EventMessage

class SubscribeEventMessage(EventMessage):

    # handler key 
    key:str = 'event.subscribe'

    def __init__(self, to_username: str, from_username: str, agent_id: str, event: str, event_key: str = '', **kwargs) -> None:
        super().__init__(to_username, from_username, agent_id, event, event_key=event_key, **kwargs)

    @classmethod
    def from_xml(cls, xml_tree: Element):
        '''Parse subscribe event message from XML object'''
        to_username:str = xml_tree.find('ToUserName').text
        from_username:str = xml_tree.find('FromUserName').text
        create_time:str = xml_tree.find('CreateTime').text
        event:str = xml_tree.find('Event').text
        event_key:str = '' if xml_tree.find('EventKey') is None else xml_tree.find('EventKey').text
        # msg_id:str = xml_tree.find('MsgId').text
        agent_id:str = xml_tree.find('AgentID').text

        return cls(to_username, from_username, agent_id, event, event_key=event_key, create_time=create_time)


#!python3.9

from . import Message
from xml.etree.ElementTree import Element

class LinkMessage(Message):

    # message type 
    type:str = 'link'
    # handler key 
    key:str = 'link'

    def __init__(self, to_username: str, from_username: str, agent_id: str, title:str, desc:str, url:str, pic_url:str, **kwargs) -> None:
        super().__init__(to_username, from_username, agent_id, **kwargs)
        # the title of the link
        self.title:str = title
        # the description of the link 
        self.description:str = desc
        # the url for the link 
        self.url:str = url
        # the url for the cover image 
        self.pic_url:str = pic_url

    def to_xml(self) -> str:
        '''Represent link message in XML format'''
        raise NotImplementedError('Unable to reply with a link message in current WeWork')

    def to_json(self) -> str:
        '''Represent link message in JSON format'''
        raise NotImplementedError('Unable to send a link message in current WeWork')
    
    @classmethod
    def from_xml(cls, xml_tree: Element):
        '''Parse link message from XML object'''
        to_username:str = xml_tree.find('ToUserName').text
        from_username:str = xml_tree.find('FromUserName').text
        create_time:str = xml_tree.find('CreateTime').text
        title:str = xml_tree.find('Title').text
        desc:str = xml_tree.find('Description').text
        url:str = xml_tree.find('Url').text
        pic_url:str = xml_tree.find('PicUrl').text
        msg_id:str = xml_tree.find('MsgId').text
        agent_id:str = xml_tree.find('AgentID').text

        return cls(to_username, from_username, agent_id, title, desc, url, pic_url, create_time=create_time, msg_id=msg_id)




#!python3.9

from .. import Message
from xml.etree.cElementTree import Element

class EventMessage(Message):

    # message type 
    type:str = 'event'
    # handler key 
    key:str = 'event'

    def __init__(self, to_username: str, from_username: str, agent_id: str, event:str, event_key:str='', **kwargs) -> None:
        super().__init__(to_username, from_username, agent_id, **kwargs)
        # event name 
        self.event:str = event
        # event key 
        self.event_key:str = event_key

    def to_xml(self) -> str:
        '''Represent event message in XML format'''
        raise NotImplementedError('Unable to reply with an event message in current WeWork')

    def to_json(self) -> str:
        '''Represent event message in JSON format'''
        raise NotImplementedError('Unable to send an event message in current WeWork')


from .view_event import ViewEventMessage
from .click_event import ClickEventMessage
from .location_event import LocationEventMessage
from .subscribe_event import SubscribeEventMessage
from .pic_weixin_event import PicWeixinEventMessage
from .enter_agent_event import EnterAgentEventMessage
from .pic_sysphoto_event import PicSysPhotoEventMessage
from .scancode_push_event import ScanCodePushEventMessage
from .location_select_event import LocationSelectEventMessage
from .batch_job_result_event import BatchJobResultEventMessage
from .scancode_waitmsg_event import ScanCodeWaitMsgEventMessage
from .pic_photo_or_album_event import PicPhotoOrAlbumEventMessage
from .share_agent_change_event import ShareAgentChangeEventMessage
from .share_chain_change_event import ShareChainChangeEventMessage
from .upload_media_job_finish_event import UploadMediaJobFinishEventMessage

def event_msg_from_xml(xml_tree:Element) -> Message:
    '''Parse message from xml tree'''
    event:str = xml_tree.find('Event').text.lower()
    if event in ['subscribe', 'unsubscribe']: return SubscribeEventMessage.from_xml(xml_tree)
    if event == 'view': return ViewEventMessage.from_xml(xml_tree)
    if event == 'click': return ClickEventMessage.from_xml(xml_tree)
    if event == 'location': return LocationEventMessage.from_xml(xml_tree)
    if event == 'pic_weixin': return PicWeixinEventMessage.from_xml(xml_tree)
    if event == 'enter_agent': return EnterAgentEventMessage.from_xml(xml_tree)
    if event == 'pic_sysphoto': return PicSysPhotoEventMessage.from_xml(xml_tree)
    if event == 'scancode_push': return ScanCodePushEventMessage.from_xml(xml_tree)
    if event == 'location_select': return LocationSelectEventMessage.from_xml(xml_tree)
    if event == 'batch_job_result': return BatchJobResultEventMessage.from_xml(xml_tree)
    if event == 'scancode_waitmsg': return ScanCodeWaitMsgEventMessage.from_xml(xml_tree)
    if event == 'pic_photo_or_album': return PicPhotoOrAlbumEventMessage.from_xml(xml_tree)
    if event == 'share_chain_change': return ShareChainChangeEventMessage.from_xml(xml_tree)
    if event == 'share_agent_change': return ShareAgentChangeEventMessage.from_xml(xml_tree)
    if event == 'upload_media_job_finish': return UploadMediaJobFinishEventMessage.from_xml(xml_tree)
    Message.logger.warn('Unregistered event type')



# -*- encoding: utf-8

import hashlib, base64, time, requests, json, random, string, struct, socket, logging, sys
import xml.etree.cElementTree as ET
from .lib import Logger
from threading import Thread
from requests import Response
from Crypto.Cipher import AES
from flask import Flask, request
from .msg import Message, msg_from_xml, TextMessage, ImageMessage, VoiceMessage, VideoMessage, LinkMessage, LocationMessage
from xml.etree.cElementTree import Element
from typing import List, Tuple, Callable, Literal, Dict, Union

log = logging.getLogger('werkzeug')
log.disabled = True
cli = sys.modules['flask.cli']
cli.show_server_banner = lambda *x: None

class WWBot:
    '''Bot class for self-built applications of WeWork'''

    logger:Logger = Logger('WWBot')
    # some api urls 
    API_PUSH_URL:str = 'https://qyapi.weixin.qq.com/cgi-bin/message/send?access_token={token}'
    # api url for sending chat messages 
    API_CHATMSG_URL:str = 'https://qyapi.weixin.qq.com/cgi-bin/appchat/send?access_token={token}'
    # api url for sending messages as group bots 
    API_WEBHOOK_URL:str = 'https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key={key}'
    API_TOKEN_URL:str = 'https://qyapi.weixin.qq.com/cgi-bin/gettoken?corpid={corpid}&corpsecret={appsecret}'

    __last_token_got_at:int = 0
    __cached_access_token:str = ""
    __access_token_lifetime:int = 7200

    max_retry:int = 3

    msg_handler:Dict[str, Callable[[Message], Message]] = {}
    __resp_msg_cache:Dict[str, Dict[Literal['nreq', 'resp'], Union[int, Message]]] = {}

    # configurations 
    corp_id:str = ''
    corp_secret:str = ''
    token:str = ''
    aes_key:bytes = b''
    callback_path:str = '/mbot'
    flask_app:Flask = None

    # message types 
    TEXT:str = 'text'
    IMAGE:str = 'image'
    VOICE:str = 'voice'
    VIDEO:str = 'video'
    LOCATION:str = 'location'
    LINK:str = 'link'

    @classmethod
    def config(cls, app:Flask, corp_id:str, corp_secret:str, token:str, aes_key:bytes, callback_path:str='/mbot'):
        '''Configure the basic arguments for the bot'''
        # flask app 
        cls.flask_app = app
        # corp id (in the page of corp. detail)
        cls.corp_id = corp_id
        # corp secret (maybe in the page of corp. detail too)
        cls.corp_secret = corp_secret
        # random token set in the page of application detail 
        cls.token = token
        # the key for AES encryption, set in the page of application detail 
        cls.aes_key = aes_key
        # the url path for the message callback 
        cls.callback_path = callback_path

        @cls.verify_handler()
        @cls.request_handler()
        def useless(): pass 

    @classmethod
    def cal_sig(cls, token:str, timestamp:str, nonce:str, enc_text:str) -> str:
        '''Calculate the SHA1 signature for the message'''
        args:List[str] = [token, timestamp, nonce, enc_text]
        args.sort()
        sig_str:str = ''.join(args)
        return hashlib.sha1(sig_str.encode('utf-8')).hexdigest()

    @classmethod
    def aes_decrypt(cls, aes_key:bytes, enc_bytes:bytes) -> bytes:
        '''Decrypt the text that encrypted with AES'''
        enc_msg:bytes = base64.b64decode(enc_bytes)
        aes:AES = AES.new(aes_key, AES.MODE_CBC, aes_key[:16])
        return aes.decrypt(enc_msg)

    @classmethod
    def aes_encrypt(cls, aes_key:bytes, text:bytes) -> str:
        '''Encrypt the text with AES'''
        aes:AES = AES.new(aes_key, AES.MODE_CBC, aes_key[:16])
        text_len:int = len(text)
        block_size:int = 32
        to_pad:int = block_size - (text_len % block_size)
        if to_pad == 0: to_pad = block_size
        return base64.b64encode(aes.encrypt(text + chr(to_pad).encode('utf-8')*to_pad)).decode('utf-8')

    @classmethod
    def get_access_token(cls, corp_id:str, corp_secret:str) -> Tuple[str, None]:
        '''Get access token'''
        now:float = time.time()
        if now - cls.__last_token_got_at < cls.__access_token_lifetime: return cls.__cached_access_token
        try:
            resp:Response = requests.get(cls.API_TOKEN_URL.format(corpid=corp_id, appsecret=corp_secret), timeout=20)
            if resp.status_code != 200:
                cls.logger.error('WxWork Receive Response With Status Code', resp.status_code)
                return None
            result:object = resp.json()
            if result['errcode'] != 0:
                cls.logger.error(result['errmsg'])
                return None
            cls.__access_token_lifetime = result['expires_in']
            cls.__last_token_got_at = now
            cls.__cached_access_token = result['access_token']
            return cls.__cached_access_token
        except Exception as e:
            cls.logger.error(e)
            return None

    @classmethod
    def send_to(cls, corp_id:str, corp_secret:str, msg:Message):
        '''Send a text message to a specific user'''
        access_token:str = cls.get_access_token(corp_id, corp_secret)
        if access_token is None:
            cls.logger.error('Failed To Get Access Token')
            return False
        url:str = cls.API_PUSH_URL.format(token=access_token)
        # if the message is for chat 
        if msg.for_chat:
            cls.logger.info('Message sent as Chat Message')
            url = cls.API_CHATMSG_URL.format(token=access_token)
        # if the message is tended to be sent as group bot 
        elif msg.for_group_bot:
            cls.logger.info('Message sent as Group Bot Message')
            url = cls.API_WEBHOOK_URL.format(key=msg.group_bot_key)
        try:
            resp:Response = requests.post(url, data=msg.to_json(), timeout=20)
            if resp.status_code != 200:
                cls.logger.error('WxWork Receive Response With Status Code', resp.status_code)
                return False
            result:object = resp.json()
            if result['errcode'] != 0:
                cls.logger.error(result['errmsg'])
                return False
            return True
        except Exception as e:
            cls.logger.error(e)
            return False

    @classmethod
    def on(cls, msg_type:str, must_reply:bool=True) -> Callable[[Callable[[Message], Message]], Callable[[Message], Message]]:
        '''Decorator for message dealing'''
        def deco(func:Callable[[Message], Message]) -> Callable[[Message], Message]:
            def on_wrapper(msg:Message) -> Message:
                if msg.message_id in cls.__resp_msg_cache and cls.__resp_msg_cache[msg.message_id]['resp'] is not None:
                    resp_msg:Message = cls.__resp_msg_cache[msg.message_id]['resp']
                    del cls.__resp_msg_cache[msg.message_id]
                    return resp_msg
                if msg.message_id in cls.__resp_msg_cache: return None

                def thread_job(msg:Message):
                    resp_msg:Message = func(msg)
                    cls.__resp_msg_cache[msg.message_id]['resp'] = resp_msg
                    if cls.__resp_msg_cache[msg.message_id]['nreq'] >= cls.max_retry:
                        if must_reply:
                            cls.logger.info(f'Reply to the message No. {msg.message_id} of {msg.from_username} asynchronously')
                            cls.send_to(cls.corp_id, cls.corp_secret, resp_msg)
                        del cls.__resp_msg_cache[msg.message_id]

                cls.__resp_msg_cache[msg.message_id] = {'nreq': 0, 'resp': None}
                thread:Thread = Thread(target=thread_job, args=(msg,))
                thread.start()
            cls.msg_handler[msg_type] = on_wrapper
            return on_wrapper
        return deco
    
    @classmethod
    def verify_handler(cls, methods:List[str]=['GET']):
        '''Decorator for url verification callback'''
        def deco(func:Callable):
            @cls.flask_app.route(cls.callback_path, methods=methods)
            def verify_handler_wrapper():
                msg_sig:str = request.args.get('msg_signature', default='')
                ts:str = request.args.get('timestamp', default='')
                nonce:str = request.args.get('nonce', default='')
                echo_str:str = request.args.get('echostr', default='')

                sig:str = cls.cal_sig(cls.token, ts, nonce, echo_str)
                if msg_sig != sig:
                    cls.logger.error('Message Signature Dismatched. Sig. Supposed: {sig}, Sig. Actual: {msg_sig}')
                    return 'Invalid Message Signature', 403
                
                dec_msg:bytes = cls.aes_decrypt(cls.aes_key, echo_str)
                rdm_str:bytes = dec_msg[:16]
                msg_len:int = int.from_bytes(dec_msg[16:20], 'big')
                msg:str = dec_msg[20:20+msg_len].decode('utf-8')

                cls.logger.info(f'Got EchoString {msg}')

                return msg, 200
            return verify_handler_wrapper
        return deco

    @classmethod
    def request_handler(cls, methods:List[str]=['POST']):
        '''Decorator for request callback'''
        def deco(func:Callable):
            @cls.flask_app.route(cls.callback_path, methods=methods)
            def request_handler_wrapper():
                msg_sig:str = request.args.get('msg_signature', default='')
                ts:str = request.args.get('timestamp', default='')
                nonce:str = request.args.get('nonce', default='')

                req_msg:str = request.get_data(as_text=True)
                
                if len(req_msg) <= 0: return 'Page not found', 404

                req_msg_xml:Element = ET.fromstring(req_msg)
                enc_msg:str = req_msg_xml.find('Encrypt').text

                sig:str = cls.cal_sig(cls.token, ts, nonce, enc_msg)
                if msg_sig != sig:
                    cls.logger.error('Message Signature Dismatched. Sig. Supposed: {sig}, Sig. Actual: {msg_sig}')
                    return 'Invalid Message Signature'
                
                dec_msg:bytes = cls.aes_decrypt(cls.aes_key, enc_msg)
                msg_len:int = int.from_bytes(dec_msg[16:20], 'big')
                msg_xml_str:str = dec_msg[20:20+msg_len].decode('utf-8')

                msg_detail_xml:Element = ET.fromstring(msg_xml_str)
                # parse message for detail 
                msg:Message = msg_from_xml(msg_detail_xml)

                if msg is None: return '', 200

                cls.logger.info(f'Got a new {msg.__class__.__name__} message, NO. {msg.message_id}')
                
                if msg.key not in cls.msg_handler:
                    cls.logger.warn(f'No message handler registered for messages of "{msg.key}" type')
                    return '', 200

                start_at:float = time.time()
                resp_msg:Message = cls.msg_handler[msg.key](msg)
                while (time.time() - start_at < 4.8) and resp_msg is None:
                    time.sleep(0.4)
                    resp_msg = cls.msg_handler[msg.key](msg)
                if resp_msg is None:
                    cls.__resp_msg_cache[msg.message_id]['nreq'] += 1
                    cls.logger.warn('Failed to reply in time, try next time')
                    time.sleep(0.2)
                    return '', 403
                cls.logger.info('Reply in time')

                resp_ts:str = str(int(time.time()))
                resp_nonce:str = ''.join(random.choices(string.ascii_letters, k=10))
                resp_rdm_str:str = ''.join(random.choices(string.ascii_letters, k=16))
                resp_recv_id:str = ''.join(random.choices(string.digits, k=16)).encode('utf-8')
                
                resp_xml:str = resp_msg.to_xml()
                resp_msg_encrypt:str = cls.aes_encrypt(cls.aes_key, resp_rdm_str.encode('utf-8') + struct.pack('I', socket.htonl(len(resp_xml.encode('utf-8')))) + resp_xml.encode('utf-8') + resp_recv_id)
                resp_msg_sig:str = cls.cal_sig(cls.token, resp_ts, resp_nonce, resp_msg_encrypt)

                return f'''<xml><Encrypt><![CDATA[{resp_msg_encrypt}]]></Encrypt><MsgSignature><![CDATA[{resp_msg_sig}]]></MsgSignature><TimeStamp>{resp_ts}</TimeStamp><Nonce><![CDATA[{resp_nonce}]]></Nonce></xml>'''
            return request_handler_wrapper
        return deco

@WWBot.on(WWBot.TEXT)
def text_default(msg:TextMessage) -> Message:
    return TextMessage(msg.from_username, msg.to_username, msg.agent_id, msg.content)

@WWBot.on(WWBot.IMAGE)
def image_default(msg:ImageMessage) -> Message:
    return ImageMessage(msg.from_username, msg.to_username, msg.agent_id, msg.media_id)

@WWBot.on(WWBot.VOICE)
def voice_defualt(msg:VoiceMessage) -> Message:
    return VoiceMessage(msg.from_username, msg.to_username, msg.agent_id, msg.media_id)

@WWBot.on(WWBot.VIDEO)
def video_default(msg:VideoMessage) -> Message:
    return VideoMessage(msg.from_username, msg.to_username, msg.agent_id, msg.media_id)

@WWBot.on(WWBot.LOCATION)
def location_default(msg:LocationMessage) -> Message:
    return TextMessage(msg.from_username, msg.to_username, msg.agent_id, f'{msg.label}:({msg.location_x},{msg.location_y})\nScale:{msg.scale}')

@WWBot.on(WWBot.LINK)
def link_default(msg:LinkMessage) -> Message:
    return TextMessage(msg.from_username, msg.to_username, msg.agent_id, f'{msg.title}\n{msg.description}\n{msg.url}\n{msg.pic_url}')



#!python3.9

import requests, string, random
from ..lib import Logger
from requests import Response
from typing import List, Dict, Any, NamedTuple

class ChatInfo(NamedTuple):
    chat_id:str
    name:str
    owner:str
    user_list:List[str]

class Chat:

    logger:Logger = Logger('Chat')

    # api url for create chat 
    API_CREATE_URL:str = 'https://qyapi.weixin.qq.com/cgi-bin/appchat/create?access_token={token}'
    # api url for update chat 
    API_UPDATE_URL:str = 'https://qyapi.weixin.qq.com/cgi-bin/appchat/update?access_token={token}'
    # api url for get chat 
    API_GET_URL:str = 'https://qyapi.weixin.qq.com/cgi-bin/appchat/get?access_token={token}&chatid={chat_id}'

    @classmethod
    def create(cls, access_token:str, name:str, user_list:List[str], owner:str=None, chat_id:str=None) -> str:
        '''Create a chat with provided user list'''
        if len(user_list) < 2: return None
        if owner is None or owner not in user_list: owner = user_list[0]
        if chat_id is None: chat_id = ''.join(random.choices(string.ascii_letters+string.digits, k=32))
        url:str = cls.API_CREATE_URL.format(token=access_token)
        try:
            resp:Response = requests.post(url, json={"name":name,"owner": owner,"userlist":user_list,"chatid":chat_id}, timeout=20)
            if resp.status_code != 200:
                cls.logger.error('WxWork Receive Response With Status Code', resp.status_code)
                return None
            result:object = resp.json()
            if result['errcode'] != 0:
                cls.logger.error(result['errmsg'])
                return None
            return result['chatid']
        except Exception as e:
            cls.logger.error(e)
            return None
    
    @classmethod
    def update(cls, access_token:str, chat_id:str, name:str=None, add_user_list:List[str]=[], del_user_list:List[str]=[], owner:str=None) -> bool:
        '''Update a chat with provided information'''
        data:Dict[str, Any] = {"chatid": chat_id}
        if name is not None: data['name'] = name
        if owner is not None: data['owner'] = owner
        if len(add_user_list) > 0: data['add_user_list'] = add_user_list
        if len(del_user_list) > 0: data['del_user_list'] = del_user_list
        url:str = cls.API_UPDATE_URL.format(token=access_token)
        try:
            resp:Response = requests.post(url, json=data, timeout=20)
            if resp.status_code != 200:
                cls.logger.error('WxWork Receive Response With Status Code', resp.status_code)
                return False
            result:object = resp.json()
            if result['errcode'] != 0:
                cls.logger.error(result['errmsg'])
                return False
            return True
        except Exception as e:
            cls.logger.error(e)
            return False
    
    @classmethod
    def get(cls, access_token:str, chat_id:str) -> ChatInfo:
        '''Get a chat with provided chat id'''
        url:str = cls.API_GET_URL.format(token=access_token, chat_id=chat_id)
        try:
            resp:Response = requests.get(url, timeout=20)
            if resp.status_code != 200:
                cls.logger.error('WxWork Receive Response With Status Code', resp.status_code)
                return None
            result:object = resp.json()
            if result['errcode'] != 0:
                cls.logger.error(result['errmsg'])
                return None
            return ChatInfo(result['chat_info']['chatid'], result['chat_info']['name'], result['chat_info']['owner'], result['chat_info']['userlist'])
        except Exception as e:
            cls.logger.error(e)
            return None



#!python3.9

from . import Message
from xml.etree.ElementTree import Element

class LocationMessage(Message):

    # message type 
    type:str = 'location'
    # handler key 
    key:str = 'location'

    def __init__(self, to_username: str, from_username: str, agent_id: str, loc_x:str, loc_y:str, scale:str, label:str, **kwargs) -> None:
        super().__init__(to_username, from_username, agent_id, **kwargs)
        # the latitude value for the location
        self.location_x:str = loc_x
        # the longitude value for the location 
        self.location_y:str = loc_y
        # the scale 
        self.scale:str = scale
        # the lable for the location 
        self.label:str = label

    def to_xml(self) -> str:
        '''Represent location message in XML format'''
        raise NotImplementedError('Unable to reply with a location message in current WeWork')

    def to_json(self) -> str:
        '''Represent location message in JSON format'''
        raise NotImplementedError('Unable to send a location message in current WeWork')
    
    @classmethod
    def from_xml(cls, xml_tree: Element):
        '''Parse location message from XML object'''
        to_username:str = xml_tree.find('ToUserName').text
        from_username:str = xml_tree.find('FromUserName').text
        create_time:str = xml_tree.find('CreateTime').text
        loc_x:str = xml_tree.find('Location_X').text
        loc_y:str = xml_tree.find('Location_Y').text
        scale:str = xml_tree.find('Scale').text
        label:str = xml_tree.find('Label').text
        msg_id:str = xml_tree.find('MsgId').text
        agent_id:str = xml_tree.find('AgentID').text

        return cls(to_username, from_username, agent_id, loc_x, loc_y, scale, label, create_time=create_time, msg_id=msg_id)




#!python3.9

import json
from . import Message
from typing import Any, Dict
from xml.etree.ElementTree import Element

class ImageMessage(Message):

    # message type 
    type:str = 'image'
    # handler key 
    key:str = 'image'

    def __init__(self, to_username: str, from_username: str, agent_id: str, media_id: str, pic_url:str = '', **kwargs) -> None:
        super().__init__(to_username, from_username, agent_id, **kwargs)
        # media id for the image 
        self.media_id:str = media_id
        # the url for the picture 
        self.pic_url:str = pic_url

    def to_xml(self) -> str:
        '''Represent image message in XML format'''
        return f'''<xml><ToUserName><![CDATA[{self.to_username}]]></ToUserName><FromUserName><![CDATA[{self.from_username}]]></FromUserName><CreateTime>{self.create_time}</CreateTime><MsgType><![CDATA[image]]></MsgType><Image><MediaId><![CDATA[{self.media_id}]]></MediaId></Image><MsgId>{self.message_id}</MsgId><AgentID>{self.agent_id}</AgentID></xml>'''

    def to_json(self) -> str:
        '''Represent image message in JSON format'''
        data:Dict[str, Any] = {"msgtype":"image","image":{"media_id":self.media_id},"safe":1 if self.safe else 0,"enable_id_trans":1 if self.enable_id_trans else 0,"enable_duplicate_check":1 if self.enable_duplicate_check else 0,"duplicate_check_interval":self.duplicate_check_interval}
        if self.for_chat: data.update({"chatid":self.chat_id})
        elif not self.for_group_bot: data.update({"touser":self.to_username,"agentid":self.agent_id})
        return json.dumps(data)
    
    @classmethod
    def from_xml(cls, xml_tree: Element):
        '''Parse image message from XML object'''
        to_username:str = xml_tree.find('ToUserName').text
        from_username:str = xml_tree.find('FromUserName').text
        create_time:str = xml_tree.find('CreateTime').text
        pic_url:str = xml_tree.find('PicUrl').text
        media_id:str = xml_tree.find('MediaId').text
        msg_id:str = xml_tree.find('MsgId').text
        agent_id:str = xml_tree.find('AgentID').text

        return cls(to_username, from_username, agent_id, media_id, pic_url=pic_url, create_time=create_time, msg_id=msg_id)




#!python3.9

import json
from . import Message
from xml.etree.ElementTree import Element
from typing import Any, Dict, List, NamedTuple

class News(NamedTuple):
    '''A piece of single news'''
    title:str
    description:str
    pic_url:str
    url:str = None
    appid:str = None
    pagepath:str = None

class NewsMessage(Message):

    # message type 
    type:str = 'news'
    # handler key 
    key:str = 'news'

    def __init__(self, to_username: str, from_username: str, agent_id: str, articles:List[News], **kwargs) -> None:
        super().__init__(to_username, from_username, agent_id, **kwargs)
        # the news list 
        self.articles:List[News] = articles

    def to_xml(self) -> str:
        '''Represent news message in XML format'''
        news_str:str = ''.join(map(lambda n:f'<item><Title><![CDATA[{n.title}]]></Title><Description><![CDATA[{n.description}]]></Description><PicUrl><![CDATA[{n.pic_url}]]></PicUrl><Url><![CDATA[{n.url}]]></Url></item>', self.articles))
        return f'''<xml><ToUserName><![CDATA[{self.to_username}]]></ToUserName><FromUserName><![CDATA[{self.from_username}]]></FromUserName><CreateTime>{self.create_time}</CreateTime><MsgType><![CDATA[news]]></MsgType><ArticleCount>{len(self.articles)}</ArticleCount><Articles>{news_str}</Articles></xml>'''

    def to_json(self) -> str:
        '''Represent news message in JSON format'''
        articles:List[Dict[str, str]] = []
        for n in self.articles:
            article:Dict[str, str] = {"title":n.title,"description":n.description}
            if n.url is not None: article["url"] = n.url
            if n.appid is not None: 
                article["appid"] = n.appid
                article["pagepath"] = n.pagepath
            articles.append(article)
        data:Dict[str, Any] = {"msgtype":"news","news":{"articles":articles},"safe":1 if self.safe else 0,"enable_id_trans":1 if self.enable_id_trans else 0,"enable_duplicate_check":1 if self.enable_duplicate_check else 0,"duplicate_check_interval":self.duplicate_check_interval}
        if self.for_chat: data.update({"chatid":self.chat_id})
        elif not self.for_group_bot: data.update({"touser":self.to_username,"agentid":self.agent_id})
        return json.dumps(data)
    
    @classmethod
    def from_xml(cls, xml_tree: Element):
        '''Parse news message from XML object'''
        raise NotImplementedError('Unable to receive and parse a news message in current WeWork')




#!python3.9
# -*- encoding: utf-8 -*-

from os import path
import sys, os, datetime, traceback, datetime

class Logger(object):
    '''Logger'''
    
    F_RED     = '\033[;31m'
    B_RED     = '\033[;41m'
    F_GREEN   = '\033[;32m'
    B_GREEN   = '\033[;42m'
    F_YELLOW  = '\033[;33m'
    B_YELLOW  = '\033[;43m'
    F_BLUE    = '\033[;34m'
    B_BLUE    = '\033[;44m'
    C_DEFAULT = '\033[0m |'

    L_INFO:int  = 1
    L_WARN:int  = 2
    L_ERROR:int = 3
    L_DEBUG:int = 4
    L_NONE:int = 5
    CLS_LEVEL:int = L_INFO

    LOG_DIR:str = 'logs'
    TOTAL_LOG_NAME:str = (datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)).strftime('%Y%m%d%H%M%S')

    def __init__(self, name:str, tz:int=8, save:bool=False, level:int=L_INFO, sym_info:str='[√]', sym_warn:str='[!]', sym_err:str='[x]'):
        self.name:str = name
        self.time_delta:datetime.timedelta = datetime.timedelta(hours=tz)
        self.save_to_file:bool = save
        self.sym_info:str = sym_info
        self.sym_warn:str = sym_warn
        self.sym_error:str = sym_err
        self.__level:int = level

    def __now_time(self):
        now = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
        return (now + self.time_delta)
    
    def __prt(self, color:str, symbol:str, *args, with_stack:bool=False, **kwarg):
        ts = self.__now_time().strftime('%H:%M:%S')
        if with_stack: call_stack:str = '->'.join(map(lambda s: f'{s[2]}:{s[1]}', traceback.extract_stack()[:-2]))
        else:
            line, caller = traceback.extract_stack()[-3][1:3]
            call_stack:str = f'[{caller}:{line}]'
        print(color + symbol, ts, f'[{self.name}]', call_stack + Logger.C_DEFAULT, *args, **kwarg)
        sys.stdout.flush()
        if self.save_to_file:
            msg = ' '.join([str(a) for a in args])
            self.__save(self.TOTAL_LOG_NAME, f'[{ts}] | {msg}{os.linesep}')
            self.__save(self.name, f'{symbol}[{ts}] | {msg}{os.linesep}')

    def __save(self, file_name:str, msg:str):
        if not path.isdir(self.LOG_DIR):
            os.mkdir(self.LOG_DIR)
            self.info(f'Directory {self.LOG_DIR} For Log Files Created.')
        log_file:str = path.join(self.LOG_DIR, f'{file_name}.log')
        with open(log_file, 'a') as outf:
            outf.write(msg)

    def clean(self):
        dirs:list = [path.join(self.LOG_DIR, d) for d in os.listdir(self.LOG_DIR)]
        files:list = [d for d in dirs if path.isfile(d) and d.endswith('.log')]
        for f in files:
            os.remove(f)
        return self

    def info(self, *args, highlight:bool=False, with_stack:bool=False, **kwarg):
        if max(self.__level, Logger.CLS_LEVEL) > Logger.L_INFO:
            return
        color = Logger.F_GREEN if not highlight else Logger.B_GREEN
        self.__prt(color, self.sym_info, *args, with_stack=with_stack, **kwarg)
        return self
    
    def dbg(self, *args, highlight:bool=False, with_stack:bool=False, **kwarg):
        if max(self.__level, Logger.CLS_LEVEL) > Logger.L_DEBUG:
            return
        color = Logger.F_BLUE if not highlight else Logger.B_BLUE
        self.__prt(color, self.sym_info, *args, with_stack=with_stack, **kwarg)
        return self

    def warn(self, *args, highlight:bool=False, with_stack:bool=False, **kwarg):
        if max(self.__level, Logger.CLS_LEVEL) > Logger.L_WARN:
            return
        color = Logger.F_YELLOW if not highlight else Logger.B_YELLOW
        self.__prt(color, self.sym_warn, *args, with_stack=with_stack, **kwarg)
        return self

    def error(self, *args, highlight:bool=False, with_stack:bool=False, **kwarg):
        if max(self.__level, Logger.CLS_LEVEL) > Logger.L_ERROR:
            return
        color = Logger.F_RED if not highlight else Logger.B_RED
        self.__prt(color, self.sym_error, *args, with_stack=with_stack, **kwarg)
        return self

    def set_level(self, v:int=0):
        if v not in range(self.L_NONE + 1):
            self.error(f'Invalid Logging Level {v}')
            return self
        self.__level = v
        return self
    
    @classmethod
    def set_global_level(cls, v:int=0):
        """
        Set Globol Logging Level Which Can Influence All The Logger
        """
        if v not in range(cls.L_NONE + 1):
            raise Exception(f'Invalid Logging Level {v}')
        cls.CLS_LEVEL = v



#!python3.9

import json
from . import Message
from typing import Any, Dict
from xml.etree.ElementTree import Element

class MarkdownMessage(Message):

    # message type 
    type:str = 'markdown'
    # handler key 
    key:str = 'markdown'

    def __init__(self, to_username: str, from_username: str, agent_id: str, content: str, **kwargs) -> None:
        super().__init__(to_username, from_username, agent_id, **kwargs)
        # the content in markdown format
        self.content:str = content

    def to_xml(self) -> str:
        '''Represent markdown message in XML format'''
        raise NotImplementedError('Unable to reply with a markdown message in current WeWork')

    def to_json(self) -> str:
        '''Represent markdown message in JSON format'''
        data:Dict[str, Any] = {"msgtype":"markdown","markdown":{"content":self.content},"safe":1 if self.safe else 0,"enable_id_trans":1 if self.enable_id_trans else 0,"enable_duplicate_check":1 if self.enable_duplicate_check else 0,"duplicate_check_interval":self.duplicate_check_interval}
        if self.for_chat: data.update({"chatid":self.chat_id})
        elif not self.for_group_bot: data.update({"touser":self.to_username,"agentid":self.agent_id})
        return json.dumps(data)
    
    @classmethod
    def from_xml(cls, xml_tree: Element):
        '''Parse markdown message from XML object'''
        raise NotImplementedError('Unable to receive and parse a markdown message in current WeWork')




#!python3.9

from . import EventMessage
from xml.etree.ElementTree import Element
from typing import List, NamedTuple

class PicInfo(NamedTuple):
    pic_md5_sum:str

class PicPhotoOrAlbumEventMessage(EventMessage):

    # handler key 
    key:str = 'event.pic_photo_or_album'

    def __init__(self, to_username: str, from_username: str, agent_id: str, event: str, count:str, pic_list:List[PicInfo], event_key: str = '', **kwargs) -> None:
        super().__init__(to_username, from_username, agent_id, event, event_key=event_key, **kwargs)
        # pic count
        self.count:str = count
        # pic list 
        self.pic_list:List[PicInfo] = pic_list

    @classmethod
    def from_xml(cls, xml_tree: Element):
        '''Parse pic_photo_or_album event message from XML object'''
        to_username:str = xml_tree.find('ToUserName').text
        from_username:str = xml_tree.find('FromUserName').text
        create_time:str = xml_tree.find('CreateTime').text
        event:str = xml_tree.find('Event').text
        event_key:str = xml_tree.find('EventKey').text
        count:str = xml_tree.find('SendPicsInfo/Count').text
        items:str = xml_tree.findall('SendPicsInfo/PicList/Item')
        pic_list:List[PicInfo] = list(map(lambda item: item.find('PicMd5Sum').text, items))
        # msg_id:str = xml_tree.find('MsgId').text
        agent_id:str = xml_tree.find('AgentID').text

        return cls(to_username, from_username, agent_id, event, count, pic_list, event_key=event_key, create_time=create_time)


#!python3.9

import json
from . import Message
from xml.etree.ElementTree import Element
from typing import Any, Dict, List, NamedTuple

class MPNews(NamedTuple):
    '''A piece of single mpnews'''
    title:str
    thumb_media_id:str
    content:str
    author:str = None
    content_source_url:str = None
    desc:str = None

class MPNewsMessage(Message):

    # message type 
    type:str = 'mpnews'
    # handler key 
    key:str = 'mpnews'

    def __init__(self, to_username: str, from_username: str, agent_id: str, articles:List[MPNews], **kwargs) -> None:
        super().__init__(to_username, from_username, agent_id, **kwargs)
        # the mpnews list 
        self.articles:List[MPNews] = articles

    def to_xml(self) -> str:
        '''Represent mpnews message in XML format'''
        raise NotImplementedError('Unable to reply with a mpnews message in current WeWork')

    def to_json(self) -> str:
        '''Represent mpnews message in JSON format'''
        articles:List[Dict[str, str]] = []
        for n in self.articles:
            article:Dict[str, str] = {"title":n.title,"thumb_media_id":n.thumb_media_id,"content":n.content}
            if n.author is not None: article["author"] = n.author
            if n.content_source_url is not None: article["content_source_url"] = n.content_source_url
            if n.desc is not None: article["digest"] = n.desc
            articles.append(article)
        data:Dict[str, Any] = {"msgtype":"mpnews","mpnews":{"articles":articles},"safe":1 if self.safe else 0,"enable_id_trans":1 if self.enable_id_trans else 0,"enable_duplicate_check":1 if self.enable_duplicate_check else 0,"duplicate_check_interval":self.duplicate_check_interval}
        if self.for_chat: data.update({"chatid":self.chat_id})
        elif not self.for_group_bot: data.update({"touser":self.to_username,"agentid":self.agent_id})
        return json.dumps(data)
    
    @classmethod
    def from_xml(cls, xml_tree: Element):
        '''Parse mpnews message from XML object'''
        raise NotImplementedError('Unable to receive and parse a mpnews message in current WeWork')





import setuptools

long_description = '''
Please refer to the homepage of the library
'''

setuptools.setup(
  name="wwbot",
  version="0.0.13b",
  author="tbbatbb",
  author_email="20682299+tbbatbb@users.noreply.github.com",
  description="A library for dealing with messages in WeWork",
  url="https://github.com/tbbatbb/WeWorkBot",
  long_description_content_type='text/markdown',
  long_description=long_description,
  packages=setuptools.find_packages(),
  classifiers=[
    "Development Status :: 2 - Pre-Alpha",
    "Framework :: Flask",
    "Intended Audience :: Developers",
    "License :: Free for non-commercial use",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3.13",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Communications :: Chat",
    "Topic :: Software Development :: Libraries :: Python Modules"
  ],
)



#!python3.9

import os, requests, mimetypes
from ..lib import Logger
from requests import Response
from requests_toolbelt import MultipartEncoder
from typing import Literal, NamedTuple, Dict, Any

mimetypes.init()

class UploadResult(NamedTuple):
    type:str
    media_id:str
    created_at:str
    err_code:int
    err_msg:str

class UploadImageResult(NamedTuple):
    url:str
    err_code:int
    err_msg:str 

class UploadByURLResult(NamedTuple):
    job_id:str
    err_code:int
    err_msg:str 

class GetUploadByURLDetail(NamedTuple):
    media_id:str
    created_at:str
    err_code:int
    err_msg:str

class GetUploadByURLResult(NamedTuple):
    detail:GetUploadByURLDetail
    status:int
    err_code:int
    err_msg:str 

class Media:
    
    logger:Logger = Logger('Media')

    @classmethod
    def __upload_file(cls, url:str, file_path:str) -> Response:
        '''Upload files to a specific url'''
        filename:str = os.path.basename(file_path)
        mimetype:str = mimetypes.guess_type(file_path)[0]
        if mimetype is None: mimetype = 'application/octet-stream'
        inf = open(file_path, 'rb')
        files = {
            'media': (
                filename,
                inf,
                mimetype
            ),
            'filename': filename,
            'Content-Disposition': 'form-data;'
        }
        form_data = MultipartEncoder(files)
        try:
            resp:Response = requests.post(url, data=form_data, headers={'Content-Type': form_data.content_type})
            inf.close()
            return resp
        except Exception as e:
            cls.logger.error(e)
            return None

    @classmethod
    def __post_req(cls, url:str, data:Dict[str, Any]) -> Response:
        '''Send POST request'''
        try:
            resp:Response = requests.post(url, json=data)
            return resp
        except Exception as e:
            cls.logger.error(e)
            return None

    @classmethod
    def upload(cls, access_token:str, file_type:Literal['image', 'voice', 'video', 'file'], file_path:str) -> UploadResult:
        '''Upload media files synchronously'''
        url:str = f'https://qyapi.weixin.qq.com/cgi-bin/media/upload?access_token={access_token}&type={file_type}'
        resp:Response = cls.__upload_file(url, file_path)
        if resp is None: return resp
        resp_json = resp.json()
        return UploadResult(resp_json['type'], resp_json['media_id'], resp_json['created_at'], resp_json['errcode'], resp_json['errmsg'])

    @classmethod
    def upload_by_url(cls, access_token:str, scene:int, type:Literal['video', 'file'], filename:str, url:str, md5:str) -> UploadByURLResult:
        '''Upload files by providing url'''
        req_url:str = f'https://qyapi.weixin.qq.com/cgi-bin/media/upload_by_url?access_token={access_token}'
        resp:Response = cls.__post_req(req_url, {
            'scene': scene,
            'type': type,
            'filename': filename,
            'url': url,
            'md5': md5
        })
        if resp is None: return resp
        resp_json = resp.json()
        return UploadByURLResult(resp_json['jobid'], resp_json['errcode'], resp_json['errmsg'])
        
    @classmethod
    def get_upload_by_url_result(cls, access_token:str, job_id:str) -> GetUploadByURLResult:
        '''Get the result for the job "upload by url"'''
        req_url:str = f'https://qyapi.weixin.qq.com/cgi-bin/media/get_upload_by_url_result?access_token={access_token}'
        resp:Response = cls.__post_req(req_url, {'jobid': job_id})
        if resp is None: return resp
        resp_json = resp.json()
        return GetUploadByURLResult(
            GetUploadByURLDetail(resp_json['detail']['media_id'], resp_json['detail']['created_at'], resp_json['detail']['errcode'], resp_json['detail']['errmsg']),
            resp_json['status'],
            resp_json['errcode'],
            resp_json['errmsg']
        )

    @classmethod
    def uploadimg(cls, access_token:str, file_path:str) -> UploadImageResult:
        '''Upload image'''
        url:str = f'https://qyapi.weixin.qq.com/cgi-bin/media/uploadimg?access_token={access_token}'
        resp:Response = cls.__upload_file(url, file_path)
        if resp is None: return resp
        resp_json = resp.json()
        return UploadImageResult(resp_json['url'], resp_json['errcode'], resp_json['errmsg'])

    @classmethod
    def download(cls, access_token:str, media_id:str, save_to:str) -> bool:
        '''Download an uploaded file with media_id'''
        url:str = f'https://qyapi.weixin.qq.com/cgi-bin/media/get?access_token={access_token}&media_id={media_id}'
        try:
            resp:Response = requests.get(url, stream=True)
            with open(save_to, 'wb') as outf:
                for iter in resp.iter_content(chunk_size=1024*32):
                    outf.write(iter)
            return True
        except Exception as e:
            cls.logger.error(e)
            return False


# -*- encoding: utf-8

import time, random
from ..lib import Logger
from abc import abstractmethod
from xml.etree.cElementTree import Element

class Message:

    logger:Logger = Logger('Message')

    # message type 
    type:str = 'message'
    # handler key 
    key:str = 'message'

    def __init__(self, to_username:str, from_username:str, agent_id:str, create_time:int=None, msg_id:int=None, safe:bool=False, enable_id_trans:bool=False, enable_duplicate_check:bool=False, duplicate_check_interval:int=1800, for_chat:bool=False, chat_id:str=None, for_group_bot:bool=False, group_bot_key:str=None) -> None:
        # the id of the receiver of the message 
        self.to_username:str = to_username
        # the id of the sender of the message 
        self.from_username:str = from_username
        # the agent id of the message
        # also known as the application id 
        self.agent_id:str = agent_id
        # timestamp for the message 
        self.create_time:int = int(create_time) if create_time is not None else int(time.time())
        # message id 
        self.message_id:int = msg_id or (self.create_time * 1000 + random.randint(1000, 9999))
        # whether the message can be shared when it's sent by send_to 
        self.safe:bool = safe
        # whether to enable the id translation
        self.enable_id_trans:bool = enable_id_trans
        # whether to check the message duplication 
        self.enable_duplicate_check:bool = enable_duplicate_check
        # the interval of message duplication checking, in second
        self.duplicate_check_interval:int = duplicate_check_interval
        # whether the message is prepared for some chat 
        self.for_chat:bool = for_chat
        # if the message is prepared for some chat, then it will be sent to the chat with chat_id 
        self.chat_id:str = chat_id
        # whether the message is sent by a group bot
        self.for_group_bot:bool = for_group_bot
        # the key to send messages as group bots
        # contained in the webhook url of the group bot, like: https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key=6449f223-4920-2394-3aa39d939ae0
        self.group_bot_key:str = group_bot_key
    
    @abstractmethod
    def to_xml(self) -> str:
        '''Represent the message in XML format'''
        pass

    @abstractmethod
    def to_json(self) -> str:
        '''Represent the message in JSON format'''
        pass

    @classmethod
    @abstractmethod
    def from_xml(cls, xml_tree:Element):
        '''Parse different types of messages from XML object'''
        pass 

from .event import event_msg_from_xml
from .text_msg import TextMessage
from .link_msg import LinkMessage
from .file_msg import FileMessage
from .image_msg import ImageMessage
from .voice_msg import VoiceMessage
from .video_msg import VideoMessage
from .news_msg import News, NewsMessage
from .location_msg import LocationMessage
from .markdown_msg import MarkdownMessage
from .textcard_msg import TextCardMessage
from .mpnews_msg import MPNews, MPNewsMessage

def msg_from_xml(xml_tree:Element) -> Message:
    '''Parse message from xml tree'''
    msg_type:str = xml_tree.find('MsgType').text
    if msg_type == 'event': return event_msg_from_xml(xml_tree)
    if msg_type == 'text': return TextMessage.from_xml(xml_tree)
    if msg_type == 'image': return ImageMessage.from_xml(xml_tree)
    if msg_type == 'voice': return VoiceMessage.from_xml(xml_tree)
    if msg_type == 'video': return VideoMessage.from_xml(xml_tree)
    if msg_type == 'link': return LinkMessage.from_xml(xml_tree)
    if msg_type == 'location': return LocationMessage.from_xml(xml_tree)
    Message.logger.warn('Unregistered message type')
